open General_helper

(*************************************************************)
(*                Simple grammar for terms                   *)
(*************************************************************)
type id = string    (* Identifiers *)

type term =
  | TmTrue
  | TmFalse
  | TmZero | TmSucc of term
  | TmVar  of id                  (* Variable *)
  | TmAbs  of id * term           (* Abstraction *)
  | TmApp  of term * term         (* Application *)

(* Match only abstractions *)
let is_abstr t = match t with
  | TmAbs _ -> true
  | _       -> false;;
(* Match reduced terms *)
let rec is_value t = match t with
  | TmAbs _ -> true
  | _       -> false;;
(* Match all 'short' terms, i.e not requiring parentheses when displayed *)
let is_simple t = match t with
  | TmVar _ -> true
  | _       -> false;;

let rec int_to_term n = if (n == 0) then TmZero
                        else TmSucc (int_to_term (n-1));;

let rec termp_to_string t = if is_simple t then term_to_string t
                            else paren (term_to_string t)
and term_to_string t = match t with
  | TmTrue        -> (vio_string "true")
  | TmFalse       -> (vio_string "false")
  | TmVar (s)     -> (blue_string s)
  | TmAbs (v,e)   -> (lambda())^" "^(blue_string v)^
                       " . "^(termp_to_string e)
  | TmApp (a,b)   -> (termp_to_string a)^" "^(termp_to_string b)


(**************************************************************)
(*           Apply a substitution v->e to a term t            *)
(**************************************************************)
let rec substitute v e t = match t with
  | TmVar (x)    -> if (x = v) then e else t
  | TmAbs (w,f)  -> if (w = v) then t (* Should alpha-rename *)
                       else TmAbs (w,substitute v e f)
  | TmApp (f,x)  -> TmApp (substitute v e f,substitute v e x);;

(**************************************************************)
(*    Alpha renames a term into another with no name clashes  *)
(* In short : rename each variable appearing in the term `t`. *)
(* The function fresh_var_name_gen can be used to provide     *)
(* fresh variables never used elsewhere. The result follows   *)
(* the Barendregt convention.                                 *)
(**************************************************************)
let rec rename t = match t with
  | TmVar _ -> t
  | TmAbs (v,e)  -> let nv = fresh_var_name_gen() in
                       TmAbs (nv,substitute v (TmVar nv) (rename e))
  | TmApp (f,x)  -> TmApp (rename f,rename x);;

(**************************************************************)
(* A context is an association list id -> term.               *)
(* The get_from_context function takes an id and returns the  *)
(* corresponding term, or an error if the variable is missing.*)
(**************************************************************)
type context = Ctx of (id * term) list
let get_from_context (Ctx l) s =
  try
    let e = List.assoc s l in rename e
  with Not_found -> failwith ("Unknown variable "^s);;

(**************************************************************)
(* Apply one step of beta reduction to a term, in a context   *)
(* The context is solely used for giving a meaning to free    *)
(* variables. It is not modified throughout reduction.        *)
(**************************************************************)
let rec reduce_one t (Ctx l) =
  match t with
  | TmVar s    -> begin try
                           let e = List.assoc s l in rename e
                    with Not_found -> failwith ("Unknown variable "^s)
                  end

  | TmAbs (v,e)  -> t

  | TmApp (f,x)     -> if (is_value f) && not(is_abstr f)
                       then failwith ("Reducing invalid application (lhs is '"^
                                        (term_to_string f)^"')")
                       else if (is_abstr f) && (is_value x)
                       then begin match f with | TmAbs (v,e) -> substitute v x e
                                          | _ -> failwith "Unreached" end
                       else if (is_abstr f) && not(is_value x)
                       then let tx = reduce_one x (Ctx l)
                            in TmApp (f, tx)
                       else let tf = reduce_one f (Ctx l)
                            in TmApp (tf, x);;

(**************************************************************)
(* Apply the beta-reduction to a term until it is a value     *)
(* This function simply applies reduce_one recursively.       *)
(* Depending on the term t, it may not terminate.             *)
(**************************************************************)
let rec reduce t ctx =
  if (is_value t) then t else reduce (reduce_one t ctx) ctx;;
